##########################################################################
# Copyright 2009 Carlos Ribeiro
#
# This file is part of Radio Tray
#
# Radio Tray is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 1 of the License, or
# (at your option) any later version.
#
# Radio Tray is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radio Tray.  If not, see <http://www.gnu.org/licenses/>.
#
##########################################################################
import sys
from urllib.request import Request as UrlRequest
from urllib.request import urlopen as urlUrlopen
from common import USER_AGENT
import xml.etree.ElementTree as ET
from io import BytesIO
import re
import logging

class AsxPlaylistDecoder:
    def __init__(self):
        self.log = logging.getLogger('upmpdcli')


    def isStreamValid(self, contentType, firstBytes):
        if ('audio/x-ms-wax' in contentType or \
            'video/x-ms-wvx' in contentType or \
            'video/x-ms-asf' in contentType or \
            'video/x-ms-wmv' in contentType) and \
            firstBytes.strip().lower().startswith(b'<asx'):
            self.log.info('Stream is readable by ASX Playlist Decoder')
            return True
        else:
            return False

        
    def extractPlaylist(self,  url):
        self.log.info('ASX: Downloading playlist...')

        req = UrlRequest(url)
        req.add_header('User-Agent', USER_AGENT)
        f = urlUrlopen(req)
        str = f.read()
        f.close()

        self.log.info('ASX: playlist downloaded, decoding...')

        try:
            root = ET.parse(BytesIO(str))
        except:
            # Last ditch: try to fix docs with mismatched tag name case
            str = re.sub('''<([A-Za-z0-9/]+)''', \
                         lambda m: "<" + m.group(1).lower(),
                         str)
            root = ET.parse(BytesIO(str))
            
        #ugly hack to normalize the XML
        for element in root.iter():
            tmp = element.tag
            element.tag = tmp.lower()

            keys = element.attrib.keys()
            for key in keys:
                element.attrib[key.lower()] = element.attrib[key]

        elts = root.findall(".//ref/[@href]")

        result = []
        for elt in elts:
            tmp = elt.attrib['href']
            if (tmp.endswith("?MSWMExt=.asf")):
                tmp = tmp.replace("http", "mms")
            result.append(tmp)

        return result



# Copyright (C) 2025,2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


def __number_to_base(n: int, b: int) -> list[int]:
    if n == 0:
        return [0]
    digits = []
    while n:
        digits.append(int(n % b))
        n //= b
    return digits[::-1]


def number_to_base_decoded(n: int, base_chars: str) -> str:
    result: str = ""
    cnv: list[int] = __number_to_base(n=n, b=len(base_chars))
    curr: int
    for curr in cnv:
        result += base_chars[curr]
    return result



# Copyright (C) 2026 J.F.Dockes
#
# License: GPL 2.1
#
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation; either version 2.1 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Auxiliary functions for the folders tree, mostly dealing with building it
import time
import shlex
import os

from recoll import recoll
from recoll import qresultstore
from recoll import rclconfig
from upmplgutils import uplog, direntry, getOptionValue
import uprclutils
import uprclinit

# All Doc fields which we may want to access (reserve slots in the
# resultstore). We use an inclusion list, and end up with a smaller
# store than by using an exclusion list, but it's a bit more difficult
# to manage.
#
# +  possibly 'xdocid' and/or 'rcludi' if/when needed?
#
_otherneededfields = [
    "albumartist",
    "allartists",
    "comment",
    "composer",
    "conductor",
    "contentgroup",
    "date",
    "dmtime",
    "discnumber",
    "embdimg",
    "filename",
    "genre",
    "group",
    "label",
    "lyricist",
    "orchestra",
    "performer",
]

# Fetch all the docs by querying Recoll with an empty query (needs recoll 1.43.10, else use
# [mime:*]), which is guaranteed to match every doc.
# This creates the main doc array, which is then used by all modules.
#
# Because we are using the resultstore, the records are not modifyable and the aliastags
# processing is performed at indexing time by rclaudio. Cf. minimtagfixer.py
def _fetchalldocs(confdir):
    start = time.time()

    rcldb = recoll.connect(confdir=confdir)
    rclq = rcldb.query()
    rclq.execute("", stemming=0)
    # rclq.execute('album:a* OR album:b* OR album:c*', stemming=0)
    uplog("Estimated alldocs query results: %d" % (rclq.rowcount))

    fields = [r[1] for r in uprclutils.upnp2rclfields.items()]
    fields += _otherneededfields
    fields += uprclinit.allMinimTags()
    fields = list(set(fields))
    #uplog(f"_fetchalldocs: store fields: {fields}")
    rcldocs = qresultstore.QResultStore()
    rcldocs.storeQuery(rclq, fieldspec=fields, isinc=True)

    end = time.time()
    uplog("Retrieved %d docs in %.2f Seconds" % (len(rcldocs), end - start))
    return rcldocs

# Create new directory entry: insert in father and append dirvec slot
# (with ".." entry)
def _createdir(dirvec, fathidx, docidx, nm):
    dirvec.append({})
    thisidx = len(dirvec) - 1
    dirvec[fathidx][nm] = (thisidx, docidx)
    dirvec[-1][".."] = (fathidx, -1)
    dirvec[-1]["."] = (thisidx, docidx)
    return len(dirvec) - 1


# Create directory for playlist. Create + populate. The docs which are pointed by the playlist
# entries may not be in the tree yet, so we don't know how to find them (can't walk the tree yet).
# Just store the diridx and populate all playlists at the end
def _createpldir(dirvec, playlists, fathidx, docidx, doc, nm):
    myidx = _createdir(dirvec, fathidx, docidx, nm)
    # We need a "." entry
    dirvec[myidx]["."] = (myidx, docidx)
    playlists.append(myidx)
    return myidx


# Compute and return the index in root of the topdir we're a child and the rest of the path
# split as a list.
# The root entry (diridx 0) is special because its keys are the
# topdirs paths, not simple names. We look with what topdir path
# this doc belongs to, then return the appropriate diridx and the
# split remainder of the path
def _splitpath(dirvec, doc):
    path = doc["url"][7:].rstrip("/")

    # Determine the root entry (topdirs element). Special because its path is not a simple
    # name. Fathidx is its index in _dirvec
    firstdiridx = -1
    for rootpath, idx in dirvec[0].items():
        if path.startswith(rootpath):
            firstdiridx = idx[0]
            break
    if firstdiridx == -1:
        # Note: this is actually common because of the recoll documents created so that artist
        # searches work (_artiststorecoll in uprcltagscreate). These don't have real existence in
        # the file system.
        #uplog(f"No parent in topdirs: {path}")
        return None, None

    # Compute rest of path. If there is none, we're not interested.
    path1 = path[len(rootpath) :]
    if len(path1) == 0:
        return None, None

    # If there is a Group field, just add it as a virtual
    # directory in the path. This only affects the visible tree,
    # not the 'real' PATHs of course.
    try:
        if doc["group"]:
            a = os.path.dirname(path1)
            b = os.path.basename(path1)
            path1 = os.path.join(a, doc["group"], b)
    except:
        pass

    # Split path. The caller will walk the list (possibly creating
    # directory entries as needed, or doing something else).
    path = path1.split("/")[1:]
    return firstdiridx, path


# Main folders build method: walk the recoll docs array and split the URLs paths to build the
# [folders] data structure
def _rcl2folders(confdir, rcldocs):
    dirvec = []
    playlists = []
    
    # We initially thought that we needed a data structure lor linking item search results to the
    # main array. Deactivated for now as it does not seem to be needed (and we would need to add
    # xdocid to the resultstore fields).
    # self._xid2idx[doc["xdocid"]] = docidx

    start = time.time()

    rclconf = rclconfig.RclConfig(confdir)
    topdirs = [os.path.expanduser(d) for d in shlex.split(rclconf.getConfParam("topdirs"))]
    topdirs = [d.rstrip("/") for d in topdirs]

    # Create the 1st entry. This is special because it holds the
    # recoll topdirs, which are paths instead of simple names. There
    # does not seem any need to build the tree between a topdir and /
    dirvec.append({})
    dirvec[0][".."] = (0, -1)
    for d in topdirs:
        dirvec.append({})
        dirvec[0][d] = (len(dirvec) - 1, -1)
        dirvec[-1][".."] = (0, -1)

    # Walk the doc list and update the directory tree according to the url: create intermediary
    # directories if needed, create leaf entry.
    for docidx in range(len(rcldocs)):
        doc = rcldocs[docidx]

        fathidx, path = _splitpath(dirvec, doc)
        if not fathidx:
            continue

        # uplog("%s"%path, file=sys.stderr)
        for idx in range(len(path)):
            elt = path[idx]
            if elt in dirvec[fathidx]:
                # This path element was already seen
                # If this is the last entry in the path, maybe update
                # the doc idx (previous entries were created for
                # intermediate elements without a Doc).
                if idx == len(path) - 1:
                    dirvec[fathidx][elt] = (dirvec[fathidx][elt][0], docidx)
                # Update fathidx for next iteration
                fathidx = dirvec[fathidx][elt][0]
            else:
                # Element has no entry in father directory (hence no
                # dirvec entry either).
                if idx != len(path) - 1:
                    # This is an intermediate element. Create a
                    # Doc-less directory
                    fathidx = _createdir(dirvec, fathidx, -1, elt)
                else:
                    # Last element. If directory, needs a dirvec entry
                    if doc["mtype"] == "inode/directory":
                        fathidx = _createdir(dirvec, fathidx, docidx, elt)
                    elif doc["mtype"] == "audio/x-mpegurl":
                        fathidx = _createpldir(dirvec, playlists, fathidx, docidx, doc, elt)
                    else:
                        dirvec[fathidx][elt] = (-1, docidx)

    if False:
        for ent in dirvec:
            uplog("%s" % ent)

    end = time.time()
    uplog("_rcl2folders took %.2f Seconds" % (end - start))
    return dirvec, playlists


# Initialize all playlists after the tree is otherwise complete
def _initplaylists(slf, confdir, rcldocs, dirvec, playlists):
    moredocs = []
    # We use a recoll db connection for creating bogus rcl docs
    rcldb = recoll.connect(confdir=confdir)
    for diridx in playlists:
        pldocidx = dirvec[diridx]["."][1]
        pldoc = rcldocs[pldocidx]
        plpath = uprclutils.docpath(pldoc)
        try:
            m3u = uprclutils.M3u(plpath)
        except Exception as ex:
            uplog(f"M3u open failed: plpath [{plpath}] : {ex}")
            continue
        for urlorpath in m3u:
            if m3u.urlRE.match(urlorpath):
                # Actual URL (usually http). Create bogus doc
                doc = uprclutils.docforurl(rcldb, urlorpath)
                moredocs.append(doc)
                docidx = len(rcldocs) + len(moredocs) - 1
                tt = doc["title"]
                dirvec[diridx][tt] = (-1, docidx)
            else:
                if not os.path.isabs(urlorpath):
                    urlorpath = os.path.join(os.path.dirname(plpath), urlorpath)
                docidx = slf.statpath(urlorpath)
                if docidx >= 0:
                    #uplog(f"Track OK for playlist [{plpath}] entry [{urlorpath}]")
                    elt = os.path.split(urlorpath)[1]
                    dirvec[diridx][elt] = (-1, docidx)
                else:
                    uplog(f"No track for playlist [{plpath}] entry [{urlorpath}]")
                    #self.statpath(urlorpath, verbose=True)
    return moredocs



# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum
from typing import Callable
from tidalapi import Quality as TidalQuality
import lafv_matcher


class PluginConstant(Enum):

    PLUGIN_RELEASE = "0.8.12"
    PLUGIN_NAME = "tidal"
    CACHED_IMAGES_DIRECTORY = "images"
    STATIC_IMAGES_DIRECTORY = "static-images"
    PLUGIN_IMAGES_DIRECTORY = "static-images"


class PluginImageDirectory(Enum):
    PAGES = "pages"
    BUTTONS = "buttons"
    GENERIC = "generic"
    TAG = "tag"


class _FunctionProxy:
    """Allow to mask a function as an Object."""
    def __init__(self, function):
        self.function = function

    def __call__(self, *args, **kwargs):
        return self.function(*args, **kwargs)


class UserAgentMatcher(Enum):
    EXACT = _FunctionProxy(lambda x, y: x == y)
    STARTS_WITH = _FunctionProxy(lambda x, y: x.startswith(y))
    ENDS_WITH = _FunctionProxy(lambda x, y: x.endswith(y))
    CONTAINS = _FunctionProxy(lambda x, y: x in y)


class UserAgentMatcherData:

    def __init__(self, user_agent_str: str, device_list: list[str], matcher: Callable[[str, str], bool]):
        self.__user_agent_str: str = user_agent_str
        self.__device_list: list[str] = device_list
        self.__matcher: Callable[[str, str], bool] = matcher

    @property
    def user_agent_str(self) -> str:
        return self.__user_agent_str

    @property
    def device_list(self) -> list[str]:
        return self.__device_list

    @property
    def matcher(self) -> Callable[[str, str], bool]:
        return self.__matcher


class UserAgentHiResWhitelist(Enum):
    KODI_21 = UserAgentMatcherData(
        "Kodi/21",
        ["Kodi 21"],
        UserAgentMatcher.STARTS_WITH.value)
    MUSIC_PLAYER_DAEMON = UserAgentMatcherData(
        "Music Player Daemon",
        ["Music Player Daemon"],
        UserAgentMatcher.STARTS_WITH.value)
    LAVF_FFPROBE = UserAgentMatcherData(
        "Lavf/ffprobe",
        ["gmrender-resurrect", "BubbleUPnp Server"],
        UserAgentMatcher.EXACT.value)
    WIIM_FIRMWARE_GE_58_76_100 = UserAgentMatcherData(
        "Lavf/58.76.100",
        ["WiiM Device"],
        lambda x, y: lafv_matcher.match(
            to_match=x,
            pattern=y,
            match_mode=lafv_matcher.LavfMatchMode.GE))

    @property
    def user_agent_str(self) -> str:
        return self.value.user_agent_str

    @property
    def device_list(self) -> list[str]:
        return self.value.device_list

    @property
    def matcher(self) -> Callable[[str, str], bool]:
        return self.value.matcher


class EnvironmentVariableName(Enum):
    UPMPD_UPNPHOSTPORT = "UPMPD_UPNPHOSTPORT"


class _ConfigParamData:

    def __init__(self, key: str, default_value: any):
        self.__key: str = key
        self.__default_value: any = default_value

    @property
    def key(self) -> str:
        return self.__key

    @property
    def default_value(self) -> any:
        return self.__default_value


class ConfigParam(Enum):

    AUDIO_QUALITY = _ConfigParamData("audioquality", TidalQuality.high_lossless)

    ALLOW_FAVORITE_ACTIONS = _ConfigParamData("allowfavoriteactions", False)
    ALLOW_BOOKMARK_ACTIONS = _ConfigParamData("allowbookmarkactions", False)
    ALLOW_STATISTICS_ACTIONS = _ConfigParamData("allowstatisticsactions", False)
    ENABLE_USER_AGENT_WHITELIST = _ConfigParamData("enableuseragentwhitelist", True)
    ENABLE_READ_STREAM_METADATA = _ConfigParamData("enablereadstreammetadata", False)
    ENABLE_DUMP_STREAM_DATA = _ConfigParamData("enabledumpstreamdata", False)
    ENABLE_CACHED_IMAGE_AGE_LIMIT = _ConfigParamData("enablecachedimageagelimit", False)
    CACHED_IMAGE_MAX_AGE_DAYS = _ConfigParamData("cachedimagemaxagedays", 60)

    TRACK_URI_ENTRY_EXPIRATION_SEC = _ConfigParamData("trackurientryexpirationsec", 240)

    TRACK_ID_REGEX = _ConfigParamData("trackidregex", "^[0-9]+$")
    VERBOSE_LOGGING = _ConfigParamData("verboselogging", False)
    SEARCH_LIMIT = _ConfigParamData("searchlimit", 15)

    ALLOW_NAMED_STATIC_IMAGES = _ConfigParamData("allownamedstaticimages", True)
    ALLOW_GENERIC_STATIC_IMAGES = _ConfigParamData("allowgenericstaticimages", True)

    # ALLOW_STATIC_IMAGES_FOR_PAGES = _ConfigParamData("allowstaticimagesforpages", True)

    SEARCH_RESULT_TRACK_AS_CONTAINER = _ConfigParamData("searchresulttrackascontainer", True)
    ALLOW_NEXT_IN_SEARCH_RESULT = _ConfigParamData("allownextinsearchresult", True)

    PAGE_ITEMS_FOR_TILE_IMAGE = _ConfigParamData("pageitemsfortileimage", 1)
    ALLOW_SEARCH_IMAGE_FOR_PAGELINK = _ConfigParamData("allowsearchimageforpagelink", False)
    ALLOW_SEARCH_IMAGE_FOR_PAGE = _ConfigParamData("allowsearchimageforpage", False)

    ENABLE_TILE_IMAGE_CACHE = _ConfigParamData("enabletileimagecache", False)

    @property
    def key(self) -> str:
        return self.value.key

    @property
    def default_value(self) -> any:
        return self.value.default_value


class ListeningQueueKey(Enum):
    ACTION_KEY = "action"
    BUTTON_TITLE_KEY = "button_title"


class Action(Enum):
    ADD = "add"
    DEL = "del"


class ListeningQueueAction(Enum):
    ADD = Action.ADD.value
    DEL = Action.DEL.value


class ActionButtonTitle(Enum):

    BOOKMARK_ADD = "Bookmarks (+)"
    BOOKMARK_RMV = "Bookmarks (-)"
    FAVORITE_ADD = "Favorites (+)"
    FAVORITE_RMV = "Favorites (-)"


listening_queue_action_add_dict: dict[str, str] = {
    ListeningQueueKey.ACTION_KEY.value: ListeningQueueAction.ADD.value,
    ListeningQueueKey.BUTTON_TITLE_KEY.value: ActionButtonTitle.BOOKMARK_ADD.value}

listening_queue_action_del_dict: dict[str, str] = {
    ListeningQueueKey.ACTION_KEY.value: ListeningQueueAction.DEL.value,
    ListeningQueueKey.BUTTON_TITLE_KEY.value: ActionButtonTitle.BOOKMARK_RMV.value}

fav_action_key: str = "action"
fav_button_title_key: str = "button_title"

fav_action_add: str = Action.ADD.value
fav_action_del: str = Action.DEL.value

fav_action_add_dict: dict[str, str] = {
    fav_action_key: fav_action_add,
    fav_button_title_key: ActionButtonTitle.FAVORITE_ADD.value}

fav_action_del_dict: dict[str, str] = {
    fav_action_key: fav_action_del,
    fav_button_title_key: ActionButtonTitle.FAVORITE_RMV.value}

fav_action_dict: dict[str, any] = {
    fav_action_add: fav_action_add_dict,
    fav_action_del: fav_action_del_dict
}

featured_type_name_playlist: str = "PLAYLIST"
tile_image_expiration_time_sec: int = 86400

oauth2_credentials_file_name: str = "oauth2.credentials.json"
pkce_credentials_file_name: str = "pkce.credentials.json"
# remove if not really used
default_max_album_tracks_per_page: int = 30

default_max_playlist_or_mix_items_per_page: int = 25

default_enable_assume_bitdepth: int = 1
default_playlist_items_per_page: int = 25
default_tracks_per_page: int = 25
default_mix_items_per_page: int = 25
default_albums_per_page: int = 25
default_artists_per_page: int = 25
default_page_items_per_page: int = 20

# this one should be disabled by default for the release
default_dump_track_to_entry_result: int = 0

default_allow_guess_stream_info_from_other_album_track: int = 1

default_max_file_age_seconds: int = 3600

default_max_get_stream_info_mix_or_playlist: int = 5

default_show_album_id: bool = True

default_badge_favorite_album: bool = True

default_titleless_single_album_view: bool = False

default_skip_non_stereo: bool = True

default_log_intermediate_url: bool = False

default_prepend_number_in_item_list: bool = False

default_serve_mode: str = "mpd"

default_listen_queue_playlist_name: str = "Listening Queue"

default_display_quality_badge: bool = False

default_enable_image_caching: bool = False

default_dump_image_caching: bool = False



#
# Copyright (C) 2019-2022 J.F.Dockes
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Manage the playlists section of the tree
#
# Object Id prefix: 0$uprcl$playlists
#
# Obect id inside the section: $p<idx> where <idx> is the document index
#  inside the global document vector.

import os, sys, subprocess

from upmplgutils import uplog, direntry, getOptionValue, getConfigObject
from uprclutils import rcldoctoentry, cmpentries
import uprclutils
import uprclinit
from recoll import recoll
import conftree
import upradioconf


class Playlists(object):
    def __init__(self, rclconfdir, rcldocs, httphp, pathprefix):
        self.rclconfdir = rclconfdir
        self._idprefix = "0$uprcl$playlists"
        self._pprefix = pathprefix
        uplog(f"PLAYLISTS pprefix {self._pprefix} idprefix {self._idprefix}")
        self._httphp = httphp
        self._rcldocs = rcldocs
        self.recoll2playlists()
        self._noradios = conftree.valToBool(getOptionValue("uprclnoradioconf"))
        if not self._noradios:
            self._radios = []
            radiolistid = self._idprefix + "$p" + str(len(self._pldocsidx))
            radios = upradioconf.UpmpdcliRadios(getConfigObject())
            for radio in radios:
                self._radios.append(upradioconf.radioToEntry(radiolistid, None, radio))

    # Return entry to be created in the top-level directory ([playlists]).
    def rootentries(self, pid):
        # The -1 entry does not count
        nlists = len(self._pldocsidx) - 1
        if not self._noradios:
            nlists += 1
        if nlists > 0:
            nlists = str(nlists) + " "
        else:
            nlists = ""
        return [
            direntry(pid + "playlists", pid, nlists + "playlists"),
        ]

    # Create the playlists static vector by filtering the global doc vector, storing the indexes of
    # the playlists.
    def recoll2playlists(self):
        # The -1 entry is because we use index 0 for our root.
        self._pldocsidx = [
            -1,
        ]
        for docidx in range(len(self._rcldocs)):
            doc = self._rcldocs[docidx]
            if doc["mtype"] == "audio/x-mpegurl":
                self._pldocsidx.append(docidx)

    # Compute index into our entries vector by 'parsing' the objid which is like
    # {_idprefix}$pPn[$e]En Pn and En are idx0 and idx1
    # idx0==0 -> our root. idx0 == len(ourlist) is valid too and does not point to an rcldoc built
    # from an actual playlist file, but to our internal radio playlist
    def _objidtoidx(self, objid):
        # uplog("playlists:objidtoidx: %s" % objid)
        if not objid.startswith(self._idprefix):
            raise Exception(f"playlists:browse: bad objid prefix in {objid}")
        # path is like $p{idx0} or $p{idx0}$e{idx1}
        path = objid[len(self._idprefix) :]
        idx1 = -1
        if not path:
            # Browsing the root.
            idx0 = 0
        else:
            if path[1] != "p":
                raise Exception(f"playlists:browse: bad objid {objid} path {path} no $p")
            epos = path.find("$e")
            if epos != -1:
                idx0 = int(path[2:epos])
                idx1 = int(path[epos + 2 :])
            else:
                idx0 = int(path[2:])
        if self._noradios:
            maxidx = len(self._pldocsidx)-1
        else:
            maxidx = len(self._pldocsidx)
        if idx0 > maxidx:
            raise Exception(f"playlists:browse: bad objid {objid} idx0 {idx0} not in range")
        return idx0, idx1

    def _idxtoentry(self, idx):
        upnpclass = "object.container.playlistContainer"
        id = self._idprefix + "$p" + str(idx)
        if not self._noradios and idx == len(self._pldocsidx):
            title = "*Upmpdcli Radios*"
        elif idx >= 1 and idx < len(self._pldocsidx):
            doc = self._rcldocs[self._pldocsidx[idx]]
            title = doc["title"] if doc["title"] else doc["filename"]
        else:
            return None
        return direntry(id, self._idprefix, title, upnpclass=upnpclass)

    # Return the contents of the playlist at index idx
    def _playlistatidx(self, idx):
        # uplog(f"playlistatidx: idx {idx}")
        rcldb = recoll.connect(confdir=self.rclconfdir)
        pldoc = self._rcldocs[self._pldocsidx[idx]]
        plpath = uprclutils.docpath(pldoc)
        folders = uprclinit.getTree("folders")
        uplog(f"playlists: tp(plpath) {type(plpath)} plpath {plpath}")
        entries = []
        try:
            m3u = uprclutils.M3u(plpath)
        except Exception as ex:
            uplog(f"M3u open failed: [{plpath}] : {ex}")
            return entries
        for urlorpath in m3u:
            if m3u.urlRE.match(urlorpath):
                # Actual URL (usually http). Create bogus doc
                doc = uprclutils.docforurl(rcldb, urlorpath)
            else:
                # Url is a path
                if not os.path.isabs(urlorpath):
                    urlorpath = os.path.join(os.path.dirname(plpath), urlorpath)
                docidx = folders.statpath(urlorpath)
                if docidx < 0:
                    uplog(f"No track found for playlist [{plpath}] entry [{urlorpath}]")
                    continue
                doc = self._rcldocs[docidx]
            pid = self._idprefix + "$p" + str(idx)
            id = pid + "$e" + str(len(entries))
            e = rcldoctoentry(id, pid, self._httphp, self._pprefix, doc)
            if e:
                entries.append(e)

        # uplog(f"playlistatidx: idx {idx} -> {entries}")
        return entries

    # Browse method
    # objid is like {_idprefix}$p<idx0> or {_idprefix}$p<idx0>$e<idx1> (meta only)
    # flag is meta or children.
    def browse(self, pid, flag, offset, count):
        uplog(f"uprclplaylists: browse: pid {pid} flag {flag}")
        idx0, idx1 = self._objidtoidx(pid)

        if flag == "meta":
            if idx0 == 0:
                return self.rootentries(self, self._idprefix)
            elif idx0 < len(self._pldocsidx):
                if idx1 == -1:
                    return [
                        self._idxtoentry(idx0),
                    ]
                else:
                    entries = self._playlistatidx(idx0)
                    if idx1 >= 0 and idx1 < len(entries):
                        return [
                            entries[idx1],
                        ]
            elif not self._noradios and idx0 == len(self._pldocsidx):
                if idx1 == -1:
                    return [
                        self._idxtoentry(idx0),
                    ]
                else:
                    if idx1 > 0 and idx1 < len(self._radios):
                        return [
                            self._radios[idx1],
                        ]
            return []

        # Browsing children
        entries = []
        if idx0 == 0:
            # Browsing root. Return contents
            # Regular playlist entries, from our doc list.
            for i in range(len(self._pldocsidx))[1:]:
                entries.append(self._idxtoentry(i))
            # Special entry for our radio list. The id is 1 beyond valid playlist ids
            if not self._noradios:
                entries.append(self._idxtoentry(len(self._pldocsidx)))
        elif not self._noradios and idx0 == len(self._pldocsidx):
            # Browsing the radio list
            entries = self._radios
        elif idx0 < len(self._pldocsidx):
            entries = self._playlistatidx(idx0)

        return entries



#!/usr/bin/env python3

"""Use this program to obtain credentials for the tidal plugin."""

# Copyright (C) 2023,2024 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import argparse
from datetime import datetime
from pathlib import Path
import tidalapi

tidal_plugin_name: str = "tidal"


def print_setting(name: str, value: str):
    """Print function for script settings."""
    print(f"{tidal_plugin_name}{name} = {value}")


auth_type_legacy: str = "legacy"
auth_type_oauth2: str = "oauth2"
auth_type_pkce: str = "pkce"

auth_type_default: str = auth_type_oauth2

auth_types: list[str] = [auth_type_legacy, auth_type_oauth2, auth_type_pkce]


def print_separator(middle_text: str, border_len: int = 20):
    """Borded text print function."""
    print(f"{'=' * border_len} {middle_text} {'=' * border_len}")


def auth_oauth2(args: argparse.Namespace, use_pkce: bool = False):
    """Performs oauth2 authentication."""
    session = tidalapi.Session()
    exists: bool = Path.exists(Path(args.f))
    if exists:
        print(f"Credentials file [{args.f}] already exists.")
    if not args.k and exists:
        # remove existing credential file
        print(f"Removing existing file [{args.f}] ...")
        Path(args.f).unlink()
        print(f"Removed existing file [{args.f}]")
    session_file = Path(args.f)
    # Will run until you complete the login process
    session.login_session_file(session_file=session_file, do_pkce=use_pkce)
    print(f"Credentials file [{args.f}] written.")
    with open(file=args.f, mode="r", encoding="utf-8") as cred_file:
        cred_dict = json.load(cred_file)
        print(f"Displaying credentials file for type [{args.f}], json format:")
        print(json.dumps(cred_dict, indent=4))


def auth_legacy(args: argparse.Namespace):
    """Performs legacy authentication (deprecated)."""
    session = tidalapi.Session()
    # Will run until you visit the printed url and link your account
    session.login_oauth_simple()
    token_type = session.token_type
    access_token = session.access_token
    refresh_token = session.refresh_token
    expiry_time = session.expiry_time
    storable_expiry_time = datetime.timestamp(expiry_time)
    cred_dict: dict[str, str] = {
        "authentication_type": "oauth2",
        "tokentype": token_type,
        "accesstoken": access_token,
        "refreshtoken": refresh_token,
        "expirytimetimestampstr": storable_expiry_time
    }
    if args.f:
        print(f"Writing the credentials to file [{args.f}] ...")
        # write a new json file
        with open(file=args.f, mode="w", encoding="utf-8") as cred_file:
            json.dump(cred_dict, cred_file, indent=4)
        print(f"Credentials written to [{args.f}]")
    print(f"Credentials file for type [{args.t}] in json format:")
    print(json.dumps(cred_dict, indent=4))


file_switch: str = "-f"
auth_type_switch: str = "-t"
keep_file_switch: str = "-k"


def main():
    """Application entrypoint."""
    parser = argparse.ArgumentParser(description='Tidal Credentials Retriever')
    parser.add_argument(
        file_switch,
        type=str,
        help='Optionally write credentials to the specified file')
    parser.add_argument(
        auth_type_switch,
        type=str,
        help='Authentication flow, OAUTH2 (default) or PKCE')
    parser.add_argument(
        keep_file_switch,
        action='store_true',
        help='Keep existing pkce file')
    args: argparse.Namespace = parser.parse_args()
    if args.f:
        print(f"Credentials will be written to [{args.f}]")
        # ensure directory exists
        dest_path: Path = Path(args.f)
        parent_path: Path = dest_path.parent.absolute() if dest_path.parent else None
        if parent_path:
            parent_path.mkdir(parents=True, exist_ok=True)
    if args.t and args.t.lower() not in auth_types:
        raise ValueError(f"Invalid authentication type [{args.t}]")
    auth_type: str = args.t if args.t else auth_type_default
    print(f"Authentication type is [{auth_type}]")
    if auth_type_legacy.lower() == auth_type.lower():
        auth_legacy(args)
    elif auth_type.lower() in [auth_type_pkce.lower(), auth_type_oauth2.lower()]:
        use_pkce: bool = auth_type.lower() == auth_type_pkce.lower()
        # a file is needed for oauth2 and pkce authentication
        if not args.f:
            args.f = ("/tmp/pkce.credentials.json"
                      if auth_type.lower() == auth_type_pkce.lower()
                      else "/tmp/oauth2.credentials.json")
            print(f"When using [{auth_type}], the [{file_switch}] switch is required, "
                  f"using a temporary file [{args.f}] ...")
        print(f"Keep existing file: [{args.k}]")
        auth_oauth2(args, use_pkce)
    else:
        raise ValueError(f"Authentication type [{auth_type}] cannot be processed")


if __name__ == "__main__":
    main()



# Copyright (C) 2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from enum import Enum


class ColumnName(Enum):
    CREATED_TIMESTAMP = "created_timestamp"
    UPDATED_TIMESTAMP = "updated_timestamp"
    ARTIST_ID = "artist_id"
    ARTIST_NAME = "artist_name"
    ARTIST_MB_ID = "artist_musicbrainz_id"
    ARTIST_ALBUM_COUNT = "artist_album_count"
    ARTIST_COVER_ART = "artist_cover_art"
    ID = "id"
    ALBUM_ID = "album_id"
    QUALITY_BADGE = "quality_badge"
    ALBUM_MB_ID = "album_musicbrainz_id"
    ARTIST_MEDIA_TYPE = "artist_media_type"
    ALBUM_ARTIST_ID = "album_artist_id"
    ALBUM_PATH = "album_path"
    ARTIST_SORT_NAME = "artist_sort_name"
    ARTIST_ROLE = "artist_role"
    ALBUM_ARTIST = "album_artist"
    ALBUM_COVER_ART = "album_cover_art"
    ALBUM_NAME = "album_name"
    ALBUM_SONG_COUNT = "album_song_count"
    ALBUM_DURATION = "album_duration"
    ALBUM_CREATED = "album_created"
    ALBUM_YEAR = "album_year"
    ALBUM_GENRE = "album_genre"
    ALBUM_USER_RATING = "album_user_rating"
    ALBUM_DISPLAY_ARTIST = "album_display_artist"
    EXPLICIT_STATUS = "explicit_status"
    ALBUM_IS_COMPILATION = "album_is_compilation"
    ALBUM_PLAY_COUNT = "album_play_count"
    ALBUM_PLAYED = "album_played"
    ALBUM_SORT_NAME = "album_sort_name"
    ALBUM_VERSION = "album_version"
    ALBUM_ORIGINAL_RELEASE_DATE_YEAR = "album_orig_rel_date_year"
    ALBUM_ORIGINAL_RELEASE_DATE_MONTH = "album_orig_rel_date_month"
    ALBUM_ORIGINAL_RELEASE_DATE_DAY = "album_orig_rel_date_day"
    ALBUM_RELEASE_DATE_YEAR = "album_rel_date_year"
    ALBUM_RELEASE_DATE_MONTH = "album_rel_date_month"
    ALBUM_RELEASE_DATE_DAY = "album_rel_date_day"
    DISC_NUM = "disc_num"
    DISC_TITLE = "disc_title"
    ALBUM_RECORD_LABEL = "album_record_label"
    ALBUM_TRACK_QUALITY_SUMMARY = "album_track_quality_summary"
    ALBUM_MOOD = "album_mood"
    ALBUM_RELEASE_TYPE = "album_release_type"
    SONG_ID = "song_id"
    SONG_ARTIST = "song_artist"
    SONG_TITLE = "song_title"
    SONG_ALBUM_ID = "song_album_id"
    SONG_ARTIST_ID = "song_artist_id"
    SONG_COMMENT = "song_comment"
    SONG_BITDEPTH = "song_bitdepth"
    SONG_BITRATE = "song_bitrate"
    SONG_CHANNEL_COUNT = "song_channel_count"
    SONG_SAMPLING_RATE = "song_sampling_rate"
    SONG_SIZE = "song_size"
    SONG_SUFFIX = "song_suffix"
    SONG_DISC_NUMBER = "song_disc_number"
    SONG_TRACK = "song_track"
    SONG_TYPE = "song_type"
    SONG_DURATION = "song_duration"
    SONG_CONTENT_TYPE = "song_content_type"
    SONG_COVER_ART = "song_cover_art"
    SONG_CREATED = "song_created"
    SONG_DISPLAY_ALBUM_ARTIST = "song_display_album_artist"
    SONG_DISPLAY_ARTIST = "song_display_artist"
    SONG_EXPLICIT_STATUS = "song_explicit_status"
    SONG_GENRE = "song_genre"
    SONG_IS_DIR = "song_is_dir"
    SONG_MEDIA_TYPE = "song_media_type"
    SONG_MUSICBRAINZ_ID = "song_musicbrainz_id"
    SONG_PATH = "song_path"
    SONG_PLAY_COUNT = "song_play_count"
    SONG_PLAYED = "song_played"
    SONG_YEAR = "song_year"
    SONG_DISPLAY_COMPOSER = "song_display_composer"
    SONG_SORT_NAME = "song_sort_name"
    SONG_ALBUM_ARTIST_ID = "song_album_artist_id"
    SONG_CONTRIBUTOR_ROLE = "song_contributor_role"
    SONG_CONTRIBUTOR_SUB_ROLE = "song_contributor_sub_role"
    SONG_LOSSLESS_STATUS = "song_lossless_status"
    ALBUM_PROPERTY_ID = "album_property_id"
    ALBUM_PROPERTY_KEY = "album_property_key"
    ALBUM_PROPERTY_VALUE = "album_property_value"
    ALBUM_AVERAGE_BITRATE = "album_average_bitrate"
    ALBUM_LOSSLESS_STATUS = "album_lossless_status"
    ALBUM_ORIGINAL_RELEASE_DATE_STR = "album_orig_rel_date_str"
    ALBUM_RELEASE_DATE_STR = "album_rel_date_str"
    ALBUM_GENRE_LIST = "album_genre_list"
    ALBUM_MOOD_LIST = "album_mood_list"
    ALBUM_RECORD_LABEL_LIST = "album_record_label_list"
    ALBUM_RELEASE_TYPE_LIST = "album_release_type_list"
    ALBUM_DISC_COUNT = "album_disc_count"
    ALBUM_MEDIA_TYPE = "album_media_type"
    SONG_STARRED = "song_starred"
    ALBUM_STARRED = "album_starred"
    ARTIST_STARRED = "artist_starred"



#!/usr/bin/env python3
from __future__ import print_function

'''
(inspired by meta-fip.py script (https://www.lesbonscomptes.com/upmpdcli))

Metadata getter for 'Radio France' internet radio stations:

 France Inter
 France Info
 France Culture
 France Bleu
 France Musique
 Fip
 Mouv'

 usage:

 upmpdcli radiolist:

 ...

 [radio '<radiostation name>']
 url          = <stream url>
 artUrl       = <url to artwork>
 metaScript   = <[path to/]filename> stationid
 preferScript = 1

 ...

 stationid = integer identifying radiostation

 example:

  [radio FIP]
  url          = https://stream.radiofrance.fr/fip/fip_hifi.m3u8?id=radiofrance
  artUrl       = http://dietpi/radio-logo/fip.png
  metaScript   = radio-france-meta.py 7
  preferScript = 1


 to do:
   include discjockey, composers, label in metadata
   include description, expressionDescription
   optimisations


  'Radio France' stationid's

  radio station			     stationid
  --------------------------------------------
  France Inter			             1
  France Info                                2
  France Culture                             5
  France Bleu Paris                         68
  France Bleu ...
  France Bleu Pays d'Auvergne               40
  France Bleu ...
  France Musique                             4
  France Musique Easy Classique            401
  France Musique Easy                      402
  France Musique Concerts Radio France     403
  France Musique Musiques du Monde Ocora   404
  France Musique La Jazz                   405
  France Musique La Contemporaine          406
  France Musique La B.O. Musiques de Films 407
  Fip                                        7
  Fip Rock                                  64
  Fip Jazz                                  65
  Fip Groove                                66
  Fip Monde                                 69
  Fip Tout Nouveau                          70
  Fip Reggae                                71
  Fip Electro                               74
  Mouv'                                      6
  Mouv' Classics                           601
  Mouv' DanceHall                          602
  Mouv' RnB & Soul                         603
  Mouv' Rap US                             604
  Mouv' Rap Français                       605
  Mouv' Kids 'n Family                     606
  Mouv' 100% Mix                            75

'''

import json
import requests
import sys
import time

def elemExists(obj, elem):
  if elem in obj:
    return True
  return False

def elemNumber(objN, elemN):
  if elemN in objN:
    return int(objN[elemN])
  return None

def elemText(objT, elemT):
  if elemT in objT:
    return str(objT[elemT])
  return None

def validArtists(a, names):
  for nm in names:
    if nm in a and a[nm]:
      return ' & '.join(a[nm]) if type((a[nm])) is list else a[nm]
  return None

def disp_title(artist, title, year, album):
  ab = ''
  yr = ''

  if year:
    yr += ' (' + str(year) + ')'

  if album:
    ab += ' [' + album + ']'

  if title and artist and year:
    dt = artist + ' – ' + title + yr + ab

  elif title and artist:
    dt = artist + ' – ' + title + ab

  elif title and year:
    dt = title + yr + ab

  elif title:
    dt = title + ab

  elif artist and album:
    dt = artist + + ' – Album: ' + ab + yr

  elif artist:
    dt = artist + yr

  elif album:
    dt = 'Album: ' + ab + yr

  return dt

def return_metadata(title, artist, album, year, artUrl, reload):
  metadata = {'title'  : display_title,
              'artist' : artist,
              'album'  : album,
              'year'   : year,
              'artUrl' : artUrl,
              'reload' : reload}
  print("%s"% json.dumps(metadata))
  return

def return_no_metadata (reload):
  metadata = {'title'  : None,
              'artist' : None,
              'album'  : None,
              'year'   : None,
              'artUrl' : None,
              'reload' : reload}
  print("%s"% json.dumps(metadata))
  return


try:
  if len(sys.argv) > 1:
    stationid = int(sys.argv[1])

  else:
    stationid = int(sys.argv)

except:
  return_no_metadata(99999999)

else:
  try:
    r = requests.get('https://api.radiofrance.fr/livemeta/pull/' + str(stationid))
    r.raise_for_status()
  except:
    return_no_metadata(99999999)

  else:
    try:
      newjsd = r.json()

    except:
      return_no_metadata(120)

    else:
      if elemExists(newjsd, 'levels'):
        metadata_levels = newjsd['levels']
        levels = []

        for level in metadata_levels:
          if elemExists(level, 'position') and elemExists(level, 'items'):

            if level['items'][int(level['position'])]:
              levels.append(level['items'][int(level['position'])])

        if elemExists(newjsd, 'steps'):
          metadata_steps = newjsd['steps']
          playing_items  = []

          for level in levels:
            playing_items.append((metadata_steps[level]))

          if playing_items:
            item      = playing_items[-1]
            embedType = elemText(item, 'embedType')
            title     = elemText(item, 'title')
            end       = elemNumber(item, 'end')
            now       = int(time.time())

            if end:
              if end > now:
                reload = end - now + 1

              else:
                reload = 10

              if embedType == 'song':
                artist        = validArtists(item, ('highlightedArtists', 'performers', 'authors'))
                year          = elemText(item, 'anneeEditionMusique')
                album         = elemText(item, 'titreAlbum')
                artUrl        = elemText(item, 'visual')
              # For radio streams title is displayed twice (in both artist and title fields),
              # in order to display artist as well, join 'title' and 'artist'
                display_title = disp_title(artist, title, year, album)

                return_metadata(display_title, artist, album, year, artUrl, reload)

              else:
                titles = []

                if len(playing_items) == 1 :
                  if elemText(playing_items[0], 'titleConcept') != elemText(playing_items[0], 'title'):
                    titles.append(elemText(playing_items[0], 'titleConcept'))
                    titles.append(elemText(playing_items[0], 'title'))
                  else:
                    titles.append(elemText(playing_items[0], 'title'))

                else:
                  for i in playing_items:
                    if elemText(i, 'title'):
                      titles.append(i['title'])

                if titles:
                 #dtitles = sorted(set(titles), key=titles.index), list(dict(dict.fromkeys(titles)) seems faster:
                  dtitles = list(dict.fromkeys(titles))
                  display_title = ' – '.join(dtitles)

                  return_metadata(display_title, None, None, None, None, reload)

                else:
                  return_no_metadata(reload)

            else:
              return_no_metadata(None)

          else:
            return_no_metadata(None)

        else:
          return_no_metadata(None)

      else:
        return_no_metadata(None)



# Copyright (C) 2023,2024,2025,2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import base64
import constants
import config
from msgproc_provider import msgproc
import persistence
from cache_type import CacheType
import sqlite3


__encoding: str = "utf-8"


def encode(data: str) -> str:
    if config.get_config_param_as_bool(constants.ConfigParam.MINIMIZE_IDENTIFIER_LENGTH):
        encoded_name: str = base64_encode(data)
        connection: sqlite3.Connection = persistence.get_working_connection(provided=None)
        kv_list: list[persistence.KeyValueItem] = persistence.get_kv_items_by_value(
            partition=CacheType.ITEM_IDENTIFIER_CODEC.cache_name,
            value=encoded_name,
            connection=connection)
        if len(kv_list) > 1:
            raise Exception(f"Duplicate entries for [{encoded_name}]")
        kv_item: persistence.KeyValueItem = kv_list[0] if len(kv_list) == 1 else None
        if kv_item:
            return kv_item.key
        else:
            persistence.lock_immediate(connection=connection)
            id_count: int = persistence.get_kv_partition_count(
                partition=CacheType.ITEM_IDENTIFIER_CODEC.cache_name,
                connection=connection)
            new_id: str = str(id_count + 1)
            persistence.save_kv_item(key_value_item=persistence.KeyValueItem(
                partition=CacheType.ITEM_IDENTIFIER_CODEC.cache_name,
                key=new_id,
                value=encoded_name),
                connection=connection,
                do_commit=True)
            return str(new_id)
    else:
        return base64_encode(data)


def decode(id: str) -> str:
    if config.get_config_param_as_bool(constants.ConfigParam.MINIMIZE_IDENTIFIER_LENGTH):
        # must be in dict, otherwise we get an exception
        encoded_kv: persistence.KeyValueItem = persistence.get_kv_item(
            partition=CacheType.ITEM_IDENTIFIER_CODEC.cache_name,
            key=id)
        if encoded_kv is None:
            raise Exception(f"codec.decode id [{id}] not found, please browse from the root of the plugin")
        # decode encoded value
        return base64_decode(encoded_kv.value)
    else:
        return base64_decode(id)


def base64_encode(v: str) -> str:
    message_bytes: bytes = v.encode(__encoding)
    base64_bytes: bytes = base64.b64encode(message_bytes)
    id: str = base64_bytes.decode(__encoding)
    return id


def base64_decode(encoded: str) -> str:
    missing_padding: int = len(encoded) % 4
    data: str = encoded
    if missing_padding != 0:
        # add missing padding characters
        data += '=' * (4 - missing_padding)
        if config.get_config_param_as_bool(constants.ConfigParam.VERBOSE_LOGGING):
            msgproc.log(f"codec.base64_decode sanitized [{data}] "
                        f"(added [{4 - missing_padding}] padding characters)")
    base64_bytes: bytes = data.encode(__encoding)
    message_bytes: bytes = base64.b64decode(base64_bytes)
    name: str = message_bytes.decode(__encoding)
    return name



# Copyright (C) 2016-2022 J.F.Dockes
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the
#   Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import print_function

import collections
import locale
import re
import os
import sys
import base64
import platform
import shlex


def _debug(s):
    print("%s" % s, file=sys.stderr)


class ConfSimple(object):
    """A ConfSimple class reads a recoll configuration file, which is
    a typical ini file (see the Recoll manual). It's a dictionary of
    dictionaries which lets you retrieve named values from the top
    level or a subsection"""

    def __init__(self, confname, tildexp=False, readonly=True, casesensitive=False):
        self.casesens = casesensitive
        if not hasattr(self, "submapcasesens"):
            self.submapcasesens = self.casesens
        self.submaps = self._makedict(True)
        self.submaps[b""] = self._makedict()
        self.subkeys_unsorted = []
        self.dotildexpand = tildexp
        self.readonly = readonly
        self.confname = confname

        try:
            f = open(confname, "rb")
        except Exception as exc:
            # _debug("Open Exception: %s" % exc)
            # File does not exist -> empty config, not an error.
            return

        self._parseinput(f)
        f.close()

    def _makedict(self, issubmaps = False):
        casesens = self.casesens
        if issubmaps:
            casesens = self.submapcasesens
        if casesens:
            return dict()
        else:
            return CaseInsensitiveDict()

    def _parseinput(self, f):
        appending = False
        line = b""
        submapkey = b""
        for cline in f:
            cline = cline.rstrip(b"\r\n")
            if appending:
                line = line + cline
            else:
                line = cline
            line = line.strip()
            if not line or line.startswith(b"#"):
                continue

            if line.endswith(b"\\"):
                line = line[:-1]
                appending = True
                continue

            appending = False
            # _debug(line)
            if line.startswith(b"["):
                line = line.strip(b"[]")
                if self.dotildexpand:
                    submapkey = os.path.expanduser(line)
                    if type(submapkey) == type(""):
                        submapkey = submapkey.encode("utf-8")
                else:
                    submapkey = line
                # _debug("Submapkey: [%s]" % submapkey)
                self.subkeys_unsorted.append(submapkey)
                continue

            nm, sep, value = line.partition(b"=")
            if not sep:
                # No equal sign in line -> considered comment
                continue

            nm = nm.strip()
            value = value.strip()
            # _debug("sk [%s] nm: [%s] value: [%s]" % (submapkey, nm, value))
            if not submapkey in self.submaps:
                self.submaps[submapkey] = self._makedict()
            self.submaps[submapkey][nm] = value

    def getSubKeys_unsorted(self):
        return [k.decode("utf-8") for k in self.subkeys_unsorted]

    def getbin(self, nm, sk=b""):
        """Returns None if not found, empty string if found empty"""
        if type(nm) != type(b"") or type(sk) != type(b""):
            raise TypeError("getbin: parameters must be binary not unicode")
        # _debug("ConfSimple::getbin nm [%s] sk [%s]" % (nm, sk))
        if not sk in self.submaps:
            return None
        if not nm in self.submaps[sk]:
            return None
        return self.submaps[sk][nm]

    def get(self, nm, sk=b"", dflt=None):
        dodecode = False
        if type(nm) == type(""):
            dodecode = True
            nm = nm.encode("utf-8")
        if type(sk) == type(""):
            sk = sk.encode("utf-8")
        # v = ConfSimple.getbin(self, nm, sk)
        v = self.getbin(nm, sk)
        if v is not None and dodecode:
            v = v.decode("utf-8")
        if v is None:
            return dflt
        return v

    def getNamesbin(self, sk=b""):
        if not sk in self.submaps:
            return None
        return list(self.submaps[sk].keys())

    def getNames(self, sk=""):
        dodecode = False
        if type(sk) == type(""):
            dodecode = True
            sk = sk.encode("utf-8")
        if not sk in self.submaps:
            return None
        names = self.getNamesbin(sk)
        if names and dodecode:
            names = [nm.decode("utf-8") for nm in names]
        return names

    def _rewrite(self):
        if self.readonly:
            raise Exception("ConfSimple is readonly")

        tname = self.confname + "-"
        f = open(tname, "wb")
        # First output null subkey submap
        if b"" in self.submaps:
            for nm, value in self.submaps[b""].items():
                f.write(nm + b"=" + value + b"\n")
        for sk, mp in self.submaps.items():
            if sk == b"":
                continue
            f.write(b"[" + sk + b"]\n")
            for nm, value in mp.items():
                f.write(nm + b"=" + value + b"\n")
        f.close()
        try:
            # os.replace works on Windows even if dst exists, but py3 only
            os.replace(tname, self.confname)
        except:
            try:
                os.rename(tname, self.confname)
            except:
                import shutil

                shutil.move(tname, self.confname)

    def setbin(self, nm, value, sk=b""):
        if self.readonly:
            raise Exception("ConfSimple is readonly")
        if sk not in self.submaps:
            self.submaps[sk] = self._makedict()
        self.submaps[sk][nm] = value
        self._rewrite()
        return True

    def set(self, nm, value, sk=b""):
        if self.readonly:
            raise Exception("ConfSimple is readonly")
        if type(nm) == type(""):
            nm = nm.encode("utf-8")
        if type(value) == type(""):
            value = value.encode("utf-8")
        if type(sk) == type(""):
            sk = sk.encode("utf-8")
        return self.setbin(nm, value, sk)


class ConfTree(ConfSimple):
    """A ConfTree adds path-hierarchical interpretation of the section keys,
    which should be '/'-separated values. When a value is requested for a
    given path, it will also be searched in the sections corresponding to
    the ancestors. E.g. get(name, '/a/b') will also look in sections '/a' and
    '/' or '' (the last 2 are equivalent)"""

    def __init__(self, *args, **kwargs):
        if platform.system() in ("Windows", "Darwin"):
            self.submapcasesens = False
        else:
            self.submapcasesens = True
        super().__init__(*args, **kwargs)
        
    def getbin(self, nm, sk=b""):
        if type(nm) != type(b"") or type(sk) != type(b""):
            raise TypeError("getbin: parameters must be binary not unicode")
        # _debug("ConfTree::getbin: nm [%s] sk [%s]" % (nm, sk))

        # Note the test for root. There does not seem to be a direct
        # way to do this in os.path
        if not sk:
            return ConfSimple.getbin(self, nm, sk)

        # Try all sk ancestors as submaps (/a/b/c-> /a/b/c, /a/b, /a, b'')
        while True:
            val = ConfSimple.getbin(self, nm, sk)
            if val is not None:
                return val
            val = ConfSimple.getbin(self, nm, sk + b"/")
            if val is not None:
                return val
            nsk = os.path.dirname(sk)
            if nsk == sk:
                # sk was already root, we're done.
                break
            sk = nsk

        return ConfSimple.getbin(self, nm)


class ConfStack(object):
    """A ConfStack manages the superposition of a list of Configuration
    objects. Values are looked for in each object from the list until found.
    This typically provides for defaults overridden by sparse values in the
    topmost file."""

    def __init__(self, nm, dirs, tp="simple", casesensitive=True):
        self.casesens = casesensitive
        fnames = []
        for dir in dirs:
            fnm = os.path.join(dir, nm)
            fnames.append(fnm)
            self._construct(tp, fnames)

    def _construct(self, tp, fnames):
        self.confs = []
        for fname in fnames:
            if tp.lower() == "simple":
                conf = ConfSimple(fname, casesensitive = self.casesens)
            else:
                conf = ConfTree(fname, casesensitive = self.casesens)
            self.confs.append(conf)

    # Accepts / returns binary strings (non-unicode)
    def getbin(self, nm, sk=b""):
        if type(nm) != type(b"") or type(sk) != type(b""):
            raise TypeError("getbin: parameters must be binary not unicode")
        for conf in self.confs:
            value = conf.getbin(nm, sk)
            if value is not None:
                return value
        return None

    def get(self, nm, sk=b""):
        dodecode = False
        if type(nm) == type(""):
            dodecode = True
            nm = nm.encode("utf-8")
        if type(sk) == type(""):
            sk = sk.encode("utf-8")
        # v = ConfSimple.getbin(self, nm, sk)
        v = self.getbin(nm, sk)
        if v is not None and dodecode:
            v = v.decode("utf-8")
        return v


# Split string of strings, with possible quoting and escaping.
# The default is do do Recoll stringToStrings emulation: whitespace
# separated, and doublequotes only (C-style). E.G.:
#   word1 word2 "compound \\"quoted\\" string" ->
#       ['word1', 'word2', 'compound "quoted string']
#
# This is not the shlex default and can be changed by setting the
# parameters
def stringToStrings(s, quotes='"', escape="\\", escapedquotes='"', whitespace=None):
    lex = shlex.shlex(s, posix=True)
    lex.whitespace_split = True
    if quotes is not None:
        lex.quotes = quotes
    if escape is not None:
        lex.escape = escape
    if escapedquotes is not None:
        lex.escapedquotes = escapedquotes
    if whitespace is not None:
        lex.whitespace = whitespace
    lex.commenters = ""
    l = []
    while True:
        tok = lex.get_token()
        if not tok:
            break
        l.append(tok)
    return l


def stringsToString(vs):
    out = []
    for s in vs:
        if (
            s.find(" ") != -1
            or s.find("\t") != -1
            or s.find("\\") != -1
            or s.find('"') != -1
        ):
            out.append('"' + s.replace("\\", "\\\\").replace('"', '\\"') + '"')
        else:
            out.append(s)
    return " ".join(out)


def valToBool(s):
    if not s:
        return False
    try:
        val = int(s)
        return val != 0
    except:
        pass
    if type(s) == type(b""):
        s = s.decode("UTF-8")
    return s[0] in "tTyY"


# CaseInsensitiveDict was copied from:
# Requests - https://github.com/psf/requests
# Copyright 2019 Kenneth Reitz
# Apache License - Version 2.0, January 2004 - http://www.apache.org/licenses

from collections.abc import MutableMapping

class CaseInsensitiveDict(MutableMapping):
    """
    A case-insensitive ``dict``-like object.

    Implements all methods and operations of
    ``collections.MutableMapping`` as well as dict's ``copy``. Also
    provides ``lower_items``.

    All keys are expected to be strings. The structure remembers the
    case of the last key to be set, and ``iter(instance)``,
    ``keys()``, ``items()``, ``iterkeys()``, and ``iteritems()``
    will contain case-sensitive keys. However, querying and contains
    testing is case insensitive:

        cid = CaseInsensitiveDict()
        cid['Accept'] = 'application/json'
        cid['aCCEPT'] == 'application/json'  # True
        list(cid) == ['Accept']  # True

    For example, ``headers['content-encoding']`` will return the
    value of a ``'Content-Encoding'`` response header, regardless
    of how the header name was originally stored.

    If the constructor, ``.update``, or equality comparison
    operations are given keys that have equal ``.lower()``s, the
    behavior is undefined.

    """

    def __init__(self, data=None, **kwargs):
        self._store = dict()
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def __setitem__(self, key, value):
        # Use the lowercased key for lookups, but store the actual
        # key alongside the value.
        self._store[key.lower()] = (key, value)

    def __getitem__(self, key):
        return self._store[key.lower()][1]

    def __delitem__(self, key):
        del self._store[key.lower()]

    def __iter__(self):
        return (casedkey for casedkey, mappedvalue in self._store.values())

    def __len__(self):
        return len(self._store)

    def lower_items(self):
        """Like iteritems(), but with all lowercase keys."""
        return ((lowerkey, keyval[1]) for (lowerkey, keyval) in self._store.items())

    def __eq__(self, other):
        if isinstance(other, collections.Mapping):
            other = CaseInsensitiveDict(other)
        else:
            return NotImplemented
        # Compare insensitively
        return dict(self.lower_items()) == dict(other.lower_items())

    # Copy is required
    def copy(self):
        return CaseInsensitiveDict(self._store.values())

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, dict(self.items()))



# Copyright (C) 2023 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from radio_station_entry import RadioStationEntry

codec_flac: str = "flac"
codec_aac: str = "aac"

mimetype_flac: str = "audio/flac"
mimetype_aac: str = "audio/aac"

title_main_mix : str = "Radio Paradise Main Mix"
title_mellow_mix : str = "Radio Paradise Mellow Mix"
title_rock_mix : str = "Radio Paradise Rock Mix"
title_global_mix : str = "Radio Paradise Global Mix"

radio_station_list: list[RadioStationEntry] = list()

def __add(codec : str, url : str, title : str, mimetype : str, bitrate : int = None):
    id : int = len(radio_station_list) + 1
    e : RadioStationEntry = RadioStationEntry(
        id = id, 
        codec = codec,
        url = url,
        title = title,
        mimetype = mimetype,
        bitrate = bitrate)
    radio_station_list.append(e)

__add(
    url="http://stream.radioparadise.com/flacm",
    title=title_main_mix,
    codec=codec_flac,
    mimetype=mimetype_flac)

__add(
    url="http://stream.radioparadise.com/aac-320",
    title=title_main_mix,
    codec=codec_aac,
    mimetype=mimetype_aac,
    bitrate=320)

__add(
    url="http://stream.radioparadise.com/aac-128",
    title=title_main_mix,
    codec=codec_aac,
    mimetype=mimetype_aac,
    bitrate=128)

__add(
    url="http://stream.radioparadise.com/mellow-flacm",
    title=title_mellow_mix,
    codec=codec_flac,
    mimetype=mimetype_flac)

__add(
    url="http://stream.radioparadise.com/mellow-320",
    title=title_mellow_mix,
    codec=codec_aac,
    mimetype=mimetype_aac,
    bitrate=320)

__add(
    url="http://stream.radioparadise.com/mellow-128",
    title=title_mellow_mix,
    codec=codec_aac,
    mimetype=mimetype_aac,
    bitrate=128)

__add(
    url="http://stream.radioparadise.com/rock-flacm",
    title=title_rock_mix,
    codec=codec_flac,
    mimetype=mimetype_flac)

__add(
    url="http://stream.radioparadise.com/rock-320",
    title=title_rock_mix,
    codec=codec_aac,
    mimetype=mimetype_aac,
    bitrate=320)

__add(
    url="http://stream.radioparadise.com/rock-128",
    title=title_rock_mix,
    codec=codec_aac,
    mimetype=mimetype_aac,
    bitrate=128)

__add(
    url="http://stream.radioparadise.com/global-flacm",
    title=title_global_mix,
    codec=codec_flac,
    mimetype=mimetype_flac)

__add(
    url="http://stream.radioparadise.com/global-320",
    title=title_global_mix,
    codec=codec_aac,
    mimetype=mimetype_aac,
    bitrate=320)

__add(
    url="http://stream.radioparadise.com/global-128",
    title=title_global_mix,
    codec=codec_aac,
    mimetype=mimetype_aac,
    bitrate=128)



import sys

APPVERSION="0.14"

# user-agent
try:
    import platform
    USER_AGENT = "%s/%s (%s %s; %s/%s (%s))" % (
            "Upmpdcli", APPVERSION, platform.system(), platform.machine(),
            platform.linux_distribution()[0], platform.linux_distribution()[1],
            platform.linux_distribution()[2])
except:
    USER_AGENT = "Upmpdcli/" + APPVERSION




# Copyright (C) 2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from album_property_metadata import AlbumPropertyMetadata
from album_property_key import AlbumPropertyKeyValue
from collections import defaultdict


class AlbumPropertyDataset:

    def __init__(self, dataset: list[AlbumPropertyMetadata]):
        self.__dataset_size: int = len(dataset)
        self.__dataset: dict[str, list[AlbumPropertyMetadata]] = defaultdict(list)
        self.__key_set: set[str] = set()
        self.__album_id_set: set[str] = set()
        self.__values_by_key: dict[str, set[str]] = defaultdict(set)
        self.__album_id_with_key: dict[str, set[str]] = defaultdict(set)
        self.__album_id_by_key_value: dict[tuple[str, str], set[str]] = defaultdict(set)
        self.__metadata_by_album_id: dict[str, list[AlbumPropertyMetadata]] = defaultdict(list)
        curr: AlbumPropertyMetadata
        for curr in dataset:
            self.__dataset[curr.album_property_key].append(curr)
            self.__album_id_set.add(curr.album_id)
            self.__album_id_with_key[curr.album_property_key].add(curr.album_id)
            self.__key_set.add(curr.album_property_key)
            self.__values_by_key[curr.album_property_key].add(curr.album_property_value)
            self.__album_id_by_key_value[(curr.album_property_key, curr.album_property_value)].add(curr.album_id)
            self.__metadata_by_album_id[curr.album_id].append(curr)

    @property
    def size(self) -> int:
        return self.__dataset_size

    @property
    def keys(self) -> list[str]:
        return list(self.__key_set)

    def get_album_id_set_for_key(self, key: str) -> set[str]:
        album_id_set_by_key: set[str] = self.__album_id_with_key[key] if key in self.__album_id_with_key else set()
        return album_id_set_by_key

    def get_album_id_count_for_key(self, key: str) -> int:
        return len(self.get_album_id_set_for_key(key=key))

    def get_values(self, key: str) -> list[str]:
        return list(self.__values_by_key[key]) if key in self.__values_by_key else []

    def get_album_id_set_by_key_value(self, key: str, value: str) -> set[str]:
        return self.__album_id_by_key_value[(key, value)] if (key, value) in self.__album_id_by_key_value else set()

    def get_album_id_list_by_key_value(self, key: str, value: str) -> list[str]:
        return list(self.get_album_id_set_by_key_value(key=key, value=value))

    def get_album_id_count_by_key_value(self, key: str, value: str) -> int:
        return len(self.__album_id_by_key_value[(key, value)]) if (key, value) in self.__album_id_by_key_value else 0

    @property
    def album_id_count(self) -> int:
        return len(self.__album_id_set)

    @property
    def album_id_set(self) -> set[str]:
        return self.__album_id_set

    @property
    def album_id_list(self) -> list[str]:
        return list(self.album_id_set)

    def _get_by_album_id(self) -> dict[str, list[AlbumPropertyMetadata]]:
        return self.__metadata_by_album_id


class AlbumPropertyDatasetProcessor:

    def __init__(
            self,
            dataset: AlbumPropertyDataset):
        self.__dataset: AlbumPropertyDataset = dataset

    def apply_filters(self, filter_list: list[AlbumPropertyKeyValue]) -> AlbumPropertyDataset:
        by_album_id: dict[str, list[AlbumPropertyMetadata]] = self.__dataset._get_by_album_id()
        # ok, now we filter.
        matching_album_id_set: set[str] = set(by_album_id.keys())
        curr_filter: AlbumPropertyKeyValue
        for curr_filter in filter_list if filter_list else []:
            # filter with key or with None?
            if curr_filter.value is None:
                # find matching.
                matching_none: set[str] = self.__dataset.album_id_set - self.__dataset.get_album_id_set_for_key(curr_filter.key)
                # calculate intersection
                matching_album_id_set = matching_album_id_set & matching_none
            else:
                matching_value: set[str] = self.__dataset.get_album_id_set_by_key_value(
                    key=curr_filter.key,
                    value=curr_filter.value)
                # calculate intersection
                matching_album_id_set = matching_album_id_set & matching_value
        # build list of matching.
        metadata_list: list[AlbumPropertyMetadata] = [item for album_id in matching_album_id_set
                                                      if album_id in by_album_id
                                                      for item in by_album_id[album_id]]
        return AlbumPropertyDataset(metadata_list)



##########################################################################
# Copyright 2009 Carlos Ribeiro
#
# This file is part of Radio Tray
#
# Radio Tray is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 1 of the License, or
# (at your option) any later version.
#
# Radio Tray is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radio Tray.  If not, see <http://www.gnu.org/licenses/>.
#
##########################################################################
import sys
import ssl
from urllib.request import Request as UrlRequest
import urllib.request, urllib.error, urllib.parse
from urllib.error import HTTPError as HTTPError
from urllib.error import URLError as URLError
from urllib.request import HTTPRedirectHandler
from http.client import BadStatusLine as BadStatusLine
from urllib.request import build_opener as urlBuild_opener
from urllib.request import HTTPSHandler
if sys.version_info < (3,5):
    def my_ssl_create_unverified_context():
        ssl_ctx = ssl.create_default_context()
        ssl_ctx.check_hostname = False
        ssl_ctx.verify_mode = ssl.CERT_NONE
        return ssl_ctx
else:
    my_ssl_create_unverified_context = ssl._create_unverified_context
from common import USER_AGENT
from PlsPlaylistDecoder import PlsPlaylistDecoder
from M3uPlaylistDecoder import M3uPlaylistDecoder
from AsxPlaylistDecoder import AsxPlaylistDecoder
from XspfPlaylistDecoder import XspfPlaylistDecoder
from AsfPlaylistDecoder import AsfPlaylistDecoder
from RamPlaylistDecoder import RamPlaylistDecoder
from UrlInfo import UrlInfo
import logging


class DummyMMSHandler(HTTPRedirectHandler):
    def __init__(self):
        self.log = logging.getLogger('upmpdcli')

    # Handle mms redirection, or let the standard code deal with it.
    def http_error_302(self, req, fp, code, msg, headers):
        #self.log.info("http_error_302: code %s headers %s" % (code, headers))
        if 'location' in headers:
            newurl = headers['location']
            if newurl.startswith('mms:'):
                raise URLError("MMS REDIRECT:" + headers["Location"])
        return HTTPRedirectHandler.http_error_302(self, req, fp, code,
                                                  msg, headers)

class StreamDecoder:

    def __init__(self, cfg_provider):
        plsDecoder = PlsPlaylistDecoder()
        m3uDecoder = M3uPlaylistDecoder()
        asxDecoder = AsxPlaylistDecoder()
        xspfDecoder = XspfPlaylistDecoder()
        asfDecoder = AsfPlaylistDecoder()
        ramDecoder = RamPlaylistDecoder()

        self.log = logging.getLogger('upmpdcli')
        
        self.decoders = [plsDecoder, asxDecoder, asfDecoder, xspfDecoder, ramDecoder, m3uDecoder]

        self.url_timeout = None

        try:
            self.url_timeout = cfg_provider.getConfigValue("url_timeout")
            if (self.url_timeout == None):
                self.log.warn("Couldn't find url_timeout configuration")
                self.url_timeout = 100
                cfg_provider.setConfigValue("url_timeout", str(self.url_timeout))
        except Exception as e:
            self.log.warn("Couldn't find url_timeout configuration")
            self.url_timeout = 100
            cfg_provider.setConfigValue("url_timeout", str(self.url_timeout))

        self.log.info('Using url timeout = %s'% str(self.url_timeout))


    def getMediaStreamInfo(self, url):
        if type(url) != type(u""):
            url = url.decode('utf-8')
        if url.startswith("http") == False:
            self.log.info('Not an HTTP url. Maybe direct stream...')
            return UrlInfo(url, False, None)

        self.log.info('Requesting stream... %s'% url)
        req = UrlRequest(url)
        req.add_header('User-Agent', USER_AGENT)

        try:
            opener = urlBuild_opener(
                DummyMMSHandler(),
                HTTPSHandler(context = my_ssl_create_unverified_context()))
            f = opener.open(req, timeout=float(self.url_timeout))
        except HTTPError as e:
            self.log.warn('HTTP Error for %s: %s' % (url, e))
            return None
        except URLError as e:
            self.log.info('URLError for %s: %s ' % (url, e))
            if str(e.reason).startswith('MMS REDIRECT'):
                newurl = e.reason.split("MMS REDIRECT:",1)[1]
                self.log.info('Found mms redirect for: %s' % newurl)
                return UrlInfo(newurl, False, None)
            else:
                return None
        except BadStatusLine as e:
            if str(e).startswith('ICY 200'):
                self.log.info('Found ICY stream')
                return UrlInfo(url, False, None)
            else:
                return None
        except Exception as e:
            print('%s: for %s: %s' % (type(e), url, e),file=sys.stderr)
            self.log.warn('%s: for %s: %s' % (type(e), url, e))
            return None

        metadata = f.info()
        firstbytes = f.read(500)
        f.close()

        try:            
            contentType = metadata["content-type"]
            self.log.info('Content-Type: %s'% contentType)
        except Exception as e:
            self.log.info("Couldn't read content-type. Maybe direct stream...")
            return UrlInfo(url, False, None)

        for decoder in self.decoders:
                
            self.log.info('Checking decoder')
            if decoder.isStreamValid(contentType, firstbytes):
                return UrlInfo(url, True, contentType, decoder)
            
        # no playlist decoder found. Maybe a direct stream
        self.log.info('No playlist decoder could handle the stream. Maybe direct stream...')
        return UrlInfo(url, False, contentType)
        

    def getPlaylist(self, urlInfo):
        return urlInfo.getDecoder().extractPlaylist(urlInfo.getUrl())

